/* 
   Ed Cashin, cashin
   CS 202, Summer 1998
   Lab: M/W - 6th
   
   Prof. E. M. Lether
   Program #4
   Due:	Monday, August 3
   
   list.h --

   This file contains the interface for the list module.  After
   calling new_list() to get a list, the user may use the function
   pointers that are fields of the head structure to access the list's
   functions.  The FREE_LIST() macro is a convenient way to get rid of
   a list once the user is done.

   The user may use the list to store data of any type.  A macro such
   as this one may be helpful for accessing the data in a node:

   #define		INT(np)		( *((int *) (np)->data_p ) )

   This definition allows the user to use a node_t pointer, "np", to
   easily refer to the data (an integer in this example) that the node
   holds.  */

#ifndef		LIST_H
#define		LIST_H

#include	<stdlib.h>

#define		FREE_LIST(hp)	destroy_list(hp); xfree(hp);


typedef
struct list_node {
  struct list_node	*prev;
  void			*data_p;
  struct list_node	*next;
} node_t;


typedef
struct list_head {
  struct list_node	*first;
  struct list_node	*last;
  size_t		data_size;

  void	(*shift)	(struct list_head *, void *to);
  void	(*unshift)	(struct list_head *, void *from);

  void	(*pop)		(struct list_head *, void *to);
  void	(*push)		(struct list_head *, void *from);

  node_t *(*ins_before)	(struct list_head *, node_t *, void *from);
  node_t *(*ins_after)	(struct list_head *, node_t *, void *from);
  void	(*del)		(struct list_head *, node_t *gonner);
} head_t;


head_t *new_list(size_t data_size);

int traverse_fwd(head_t *hp, void (*do_something) (void *));
int traverse_backwd(head_t *hp, void (*do_something) (void *));

int destroy_list(head_t *hp);

#endif		/* no LIST_H */
