/* 
   Ed Cashin, cashin
   CS 202, Summer 1998
   Lab: M/W - 6th
   
   Prof. E. M. Lether
   Program #3:	Section 13.1, Programming Project 2 (Shiflet), p.801
   Due:	Tuesday, July 21
   Late Due Date: Sunday, July 26
   
   vector.c --- self-growing arrays

   (This file was written June 21, before the assignment, but is very
   handy in the crossword assignment for remembering the positions in
   the grid where words can go.)

   Shiflet suggests making a program to sort words.  

   I thought, "If the words are of an unknown number, you'd want an
   array that can grow." I figured that by making it general, it might
   come in handy for other things too.  The name is from Stroustrup,
   who calls his growing arrays "vectors".

   But although the user gets to use the vector in a manner that looks
   a lot like C++, this is just C, only things covered in Shiflet.

   Usage: 
   To use a vector, first create one by calling the new_vec()
   function like this...

      #include  "vector.h"

      vec_t *vec = new_vec(sizeof(char));

   ...where the type inside sizeof is the kind you want the vector to
   hold.  Once you've got a vector, you can access its methods through
   its function pointers like this...

      char 	ch = 'c';
      char 	*ch_p;

      vec->add(vec, &ch);
      ch_p = vec->index(vec, 0);   
      printf("first element in vector is: %c\n", *ch_p);
      vec->destroy(vec);
      xfree(vec);

   Requires xmalloc.c (also today's -- Saturday, June 21) which is a
   bunch of wrappers for the malloc-type functions that encorporate
   error checking.
   
   News: 

   7/17/98 -- It seems that character vectors are handy enough
   to merit some nice functions made especially for them.  In
   vector.c, "chv" is the abbreviation for character vector.  I've
   attempted to imitate the string.h functions and provide similar
   names.

   Also improved the test in add_elem().
 */


#include	<stdlib.h>
#include	<stdio.h>
#include	"vector.h"
#include	"xmalloc.h"


/* Increase DELTA_DEFAULT if you anticipate needing a lot of
   elements. */
#define		DELTA_DEFAULT		128
#define		THIS_FUNC		__FUNCTION__


/* Prototypes only needed in this file. */
void add_elem(vec_t *vec, const void *elem);
void *index_vec(vec_t *vec, int idx);
vec_t *chvec_cpycat(vec_t *chv, const char *str, int n, 
		    int clear_vec);
vec_t *chv_cpycat(vec_t *chv, const char *str, int clear_vec);


/* This function accepts a pointer to a vector and a pointer to the
   element that needs to be stored.

   It copies an element into the vector, allocating more space if
   necessary.

   There is nary a return value.
 */
void add_elem(vec_t *vec, const void *elem)
{
    size_t	used_bytes = vec->n_elems * vec->elem_size;
    
    if(used_bytes + vec->elem_size > vec->alloced_mem){
      /* need more memory for elements */
      vec->alloced_mem += vec->delta;
      vec->vec_start = xrealloc(vec->vec_start, 
				     vec->alloced_mem, 
				THIS_FUNC);
    }

    memcpy(vec->vec_start + vec->n_elems * vec->elem_size, 
	   elem, vec->elem_size);

    ++vec->n_elems;
}


/* This function receives a pointer to a vector.

   As far as the user's concerned, it could do all kinds of stuff, and
   one day it may, but for now it just frees the array space.

   There is not a return value.
 */
void destroy_vec(vec_t *vec)
{
    if (vec->vec_start)
      xfree(vec->vec_start);
}


/* This function takes a pointer to a vector.

   It resets the array, but doesn't free any memory.

   And the selfish little thing doesn't give anything back. */
void clear_vec(vec_t *vec)
{
    vec->n_elems = 0;
}


/* This function accepts a pointer to a vector and an integer that
   serves as an indicator of the offset of the element that the caller
   wants access to.

   This function allows the user to index the vector as if it were an
   array.  An important difference, though, is that the return type is
   a pointer to the element, and not the value of the element.

   And does it give a return value to the caller?  No.  
 */
void *index_vec(vec_t *vec, int idx)
{
    void *data;
    data = vec->vec_start + idx * vec->elem_size;
    return data;
}


/* Here's something different: this function only accepts a size_t--no
   pointer to a vector.

   It initializes everything in a new vec_t object and returns a
   pointer to it, based on the elem_size parameter.  */
vec_t *new_vec(size_t elem_size)
{
    vec_t *new_vec_p = xmalloc(sizeof(vec_t));

    new_vec_p->n_elems 		= 0;
    new_vec_p->elem_size 	= elem_size;
    new_vec_p->alloced_mem 	= 0;

    /* if the default is smaller than the element size, use three the
       size of three elements as the delta */
    new_vec_p->delta		= 
      (DELTA_DEFAULT > elem_size) ? DELTA_DEFAULT : (elem_size * 3);

    new_vec_p->add 		= add_elem;
    new_vec_p->index 		= index_vec;
    new_vec_p->destroy 		= destroy_vec;
    new_vec_p->clear 		= clear_vec;
    new_vec_p->vec_start	= NULL;

    return new_vec_p;
}


/* Accepts a pointer to a vector and a null-terminated character array
   that it won't change.  Also accepts an integer showing how many
   characters to copy onto the end (unless it reaches null first.)

   The character-vector analog of strncat.  Calls chvec_cpycat, which
   does the real work.

   Returns a pointer to the vector or null on error. */
vec_t *chvec_cat(vec_t *chv, const char *str, int n)
{
    /* zero means "don't clear the vector" */
    return chvec_cpycat(chv, str, n, 0);
}


/* Accepts a pointer to a vector and a null-terminated character array
   that it won't change.  Also accepts an integer showing how many
   characters to copy onto the beginning (unless it reaches null first.)

   The character-vector analog of strncpy.  Calls chvec_cpycat, which
   does the real work. 

   Returns a pointer to the vector or null on error. */
vec_t *chvec_cpy(vec_t *chv, const char *str, int n)
{
    /* one means "clear the vector" */
    return chvec_cpycat(chv, str, n, 1);
}


/* Accepts same parameters as the above functions (I am going to
   complain now.  I don't know whether anyone will ever read this, but I
   think that it is excessive to have comments for every parameter in
   every function and for the return value when the same thing is stated
   right below in C.)  

   Anyway, it also accepts an additional integer that indicates
   whether to clear the vector before copying stuff into it.

   Acts like strncpy or strncat depending on the "clear_vec"
   parameter.  If it's non-zero, it clears the vector first (like
   strncpy does for strings.)

   Returns a pointer to the vector or NULL on error.
  */
vec_t *chvec_cpycat(vec_t *chv, const char *str, int n, 
		    int clear_vec)
{
    if(chv->elem_size != sizeof(char))
      return NULL;
    if(clear_vec)
      chv->clear(chv);

    while(n-- && *str)
      chv->add(chv, str++);

    return chv;
}


/* Accepts a pointer to a character vector and returns a pointer to a
   dynamically allocated string whose contents have been copied from
   the character vector.

   It's the caller's responsibility to free the memory pointed to by
   the address in the return value.  */
char *chv2str(vec_t *chv)
{
    int		i;
    char 	*str;
    
    if(chv->elem_size != sizeof(char))
      return NULL;

    str = xmalloc(sizeof(char) * (chv->n_elems + 1));

    for(i = 0; i < chv->n_elems; ++i)
      str[i] = *((char *) chv->index(chv, i));
    str[i] = '\0';

    return str;
}


/* Accepts pointer to vector and to string.

   The character-vector analog of strcat.  Calls chv_cpycat, which
   does the real work. 

   Returns pointer to vector or NULL on error.
*/
vec_t *chv_cat(vec_t *chv, const char *str)
{
    /* zero means "don't clear the vector" */
    return chv_cpycat(chv, str, 0);
}


/* Accepts pointer to vector and to string.

   The character-vector analog of strcpy.  Calls chv_cpycat, which
   does the real work. 

   Returns pointer to vector or NULL on error.
*/
vec_t *chv_cpy(vec_t *chv, const char *str)
{
    /* one means "clear the vector" */
    return chv_cpycat(chv, str, 1);
}


/* Accepts pointer to vector and to string, along with additional
   integer that says whether to clear this vector before copying stuff
   into it.

   This one is like strcpy or strcat, but copying to a character
   vector, and not copying the terminal null character.

   It acts like strcpy if the "clear_vec" parameter is non-zero, like
   strcat otherwise.

   It returns the SAME THING AS ALL THESE ... OK.  Calm.

   It returns a pointer to the vector or NULL on error.
  */
vec_t *chv_cpycat(vec_t *chv, const char *str, int clear_vec)
{
    if(chv->elem_size != sizeof(char))
      return NULL;
    if(clear_vec)
      chv->clear(chv);

    for( ; *str; ++str)
      /* Add each character pointed to by line marker. */
      chv->add(chv, str);

    return chv;
}


/* Accepts pointer to vec_t and pointer to a filler character and an
   integer showing how many characters to put into the vector.

   This is a bit like memset, I guess.  It puts "n" "*filler_p"'s into
   "chv", and it doesn't return a thang. */
void chv_set(vec_t *chv, char *filler_p, int n)
{
    if(chv->elem_size != sizeof(char))
      return;

    for( ; n; --n)
      chv->add(chv, filler_p);    
}


/* Accepts pointer to character vector.

   This is just a convenient wrapper for fprint_chv, analogous to
   printf.  It accepts a pointer to a character vector and passes
   along the return value of fprint_chv.
 */
int print_chv(vec_t *chv)
{
    return fprint_chv(stdout, chv);
}


/* Accepts pointer to a FILE and to a character vector.

   This function prints the contents of the character vector, "chv",
   to the output file pointed to by "ofile".  It returns the number of
   characters printed. 
*/
int fprint_chv(FILE *ofile, vec_t *chv)
{
    int 	ch_printed = 0, i;

    if(!chv 
       || (chv->elem_size != sizeof(char))
       || !ofile
       )
      return 0;
    
    for(i = 0; i < chv->n_elems; ++i){
      putc( *((char *)chv->index(chv, i)),
	   ofile);
      ++ch_printed;
    }

    return ch_printed;
}

/* done. */
