/* 
   Ed Cashin, cashin
   CS 202, Summer 1998
   Lab: M/W - 6th
   
   Prof. E. M. Lether
   Program #4
   Due:	Monday, August 3
  
   xmalloc.c -- 
   Contains function definitions for simple wrappers for the
   malloc-type functions, adding error-checking code, and calling
   exit() in case of a problem after printing a simple message to the
   standard output.

   (This file was written June 21, before the assignment,
   but is handy for calling malloc-type functions without having to
   type error checking every time.)

   News: 

   7/16/98 -- added xstrdup.  Why?  The strdup function may be in
   string.h, but it is the same as these other functions in that it
   uses malloc to get new memory and returns NULL on failure.  */


#include        <stdio.h>
#include        <string.h>
#include        <stdlib.h>
#include	"xmalloc.h"


/* Calls xmalloc and checks for a NULL return value.  Prints an error
   message to stdout and exits on error.  

   Same parameters and return value as its namesake. */
void *xmalloc(size_t bytes)
{
    void *new_mem = malloc(bytes);

    if(new_mem == NULL){
      fprintf(stderr, "Virtual memory exhausted.\nExiting.\n");
      exit(EXIT_FAILURE);
    }
    else
      return new_mem;
}


/* The only thing surprising about this function is that it has an
   extra parameter for a string that contains the name of the caller
   function, so it will tell the user where the error occurred.  

   Same parameters and return value as its namesake, with the addition
   of an extra parameter for the caller's name.  The caller can use
   THIS_FUNCTION (#define'd in the header) for that purpose. */
void *xrealloc(void *mem, size_t bytes, char *caller)
{
    void *new_mem  = realloc(mem, bytes);

    if(new_mem == NULL){
      fprintf(stderr, "Virtual memory exhausted in function: "
	      "%s\nExiting.\n", caller);
      exit(EXIT_FAILURE);
    }
    else
      return new_mem;
}


/* Accepts a pointer to a string which it won't change.

   Calls strdup, checks for error, exiting on failure, and returns a
   character pointer to the new copy of the string.  */
char *xstrdup(const char*str)
{
    char *new_str = strdup(str);

    if(new_str == NULL){
      fprintf(stderr, "Virtual memory exhausted.\nExiting.\n");
      exit(EXIT_FAILURE);
    }
    else
      return new_str;
}


/* Won't try to free a NULL pointer. 

   Same parameters and return value as its namesake. */
void xfree(void *old_mem)
{
    if(old_mem) 
      free(old_mem);
    else
      fprintf(stderr, "Attempt to free NULL pointer in %s.\n",
	      THIS_FUNC);
}


/* Program will exit if the ptr is NULL. 

   Accepts a void pointer to test against NULL and a character pointer
   to a string indicating the name of the caller.  Just for asserting
   that any pointer is non-null.

   No return value.  */
void nullcheck(void *ptr, char *func)
{
    if(ptr) 
      return;

    fprintf(stderr, "NULL pointer error in %s\n", func);
    exit(EXIT_FAILURE);
}

